/*******************************************************************************
 * Copyright (c) 2009, 2017 IBM Corporation and others.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * Contributors:
 *     IBM Corporation - initial API and implementation
 *     Zend Technologies
 *     Dawid Pakuła [515715]
 *******************************************************************************/
package org.eclipse.php.ui.text.correction;

import org.eclipse.core.runtime.CoreException;
import org.eclipse.dltk.compiler.problem.IProblemIdentifier;
import org.eclipse.dltk.core.ISourceModule;
import org.eclipse.dltk.ui.text.completion.IScriptCompletionProposal;

/**
 * Interface to be implemented by contributors to the extension point
 * <code>org.eclipse.php.ui.quickFixProcessors</code>.
 * <p>
 * Since 3.2, each extension specifies the marker types it can handle, and
 * {@link #hasCorrections(ISourceModule, IProblemIdentifier)} and
 * {@link #getCorrections(IInvocationContext, IProblemLocation[])} are called if
 * (and only if) quick fix is required for a problem of these types.
 * </p>
 * <p>
 * Note, if a extension does not specify marker types it will be only called for
 * problem of type <code>org.eclipse.dltk.core.problem</code>,
 * <code>org.eclipse.dltk.core.buildpath_problem</code> and
 * <code>org.eclipse.dltk.core.task</code>; compatible with the behavior prior
 * to 3.2
 * </p>
 * 
 * @since 3.0
 */
public interface IQuickFixProcessor {

	/**
	 * Returns <code>true</code> if the processor has proposals for the given
	 * problem. This test should be an optimistic guess and be very cheap.
	 * 
	 * @param unit
	 *            the compilation unit
	 * @param identifier
	 *            the problem Id. The id is of a problem of the problem type(s) this
	 *            processor specified in the extension point.
	 * @return <code>true</code> if the processor has proposals for the given
	 *         problem
	 * @since 5.0
	 */
	public boolean hasCorrections(ISourceModule unit, IProblemIdentifier identifier);

	/**
	 * Collects corrections or code manipulations for the given context.
	 * 
	 * @param context
	 *            Defines current compilation unit, position and a shared AST
	 * @param locations
	 *            Problems are the current location.
	 * @return the corrections applicable at the location or <code>null</code> if no
	 *         proposals can be offered
	 * @throws CoreException
	 *             CoreException can be thrown if the operation fails
	 */
	IScriptCompletionProposal[] getCorrections(IInvocationContext context, IProblemLocation[] locations)
			throws CoreException;

}
